import hashlib
import hmac
import math
import requests

# thanks to https://github.com/glomatico/votify/blob/main/votify/totp.py
class TOTP:
    def __init__(self) -> None:
        self.secret, self.version = self.get_secret_version()
        self.period = 30
        self.digits = 6

    def generate(self, timestamp: int) -> str:
        counter = math.floor(timestamp / 1000 / self.period)
        counter_bytes = counter.to_bytes(8, byteorder="big")

        h = hmac.new(self.secret, counter_bytes, hashlib.sha1)
        hmac_result = h.digest()

        offset = hmac_result[-1] & 0x0F
        binary = (
            (hmac_result[offset] & 0x7F) << 24
            | (hmac_result[offset + 1] & 0xFF) << 16
            | (hmac_result[offset + 2] & 0xFF) << 8
            | (hmac_result[offset + 3] & 0xFF)
        )

        return str(binary % (10**self.digits)).zfill(self.digits)
    
    def get_secret_version(self) -> tuple[str, int]:
        req = requests.get("https://raw.githubusercontent.com/Thereallo1026/spotify-secrets/refs/heads/main/secrets/secrets.json")
        if req.status_code != 200:
            raise ValueError("Failed to fetch TOTP secret and version.")
        data = req.json()[-1]
        ascii_codes = [ord(c) for c in data['secret']]
        transformed = [val ^ ((i % 33) + 9) for i, val in enumerate(ascii_codes)]
        secret_key = "".join(str(num) for num in transformed)
        return bytes(secret_key, 'utf-8'), data['version']


import argparse
import json
import os
import subprocess
import sys

parser = argparse.ArgumentParser()

parser.add_argument("-d",
                    "--directory",
                    metavar="PATH",
                    help='give path to download directory.'
                    )

parser.add_argument("-f",
                    "--force",
                    metavar="BOOLEAN",
                    default=True,
                    help='skip check for if it already downloaded and is available in directory. '
                    )

parser.add_argument("-c",
                    "--config",
                    nargs='?',
                    const="edit",
                    help='edit config file or reset it.'
                    )

parser.add_argument("-u",
                    "--user",
                    nargs='?',
                    const="current",
                    help="items to download from users account like playlist, tracks etc.",
                    choices=['current', 'album', 'play']
                    )

parser.add_argument("URL",
                    metavar="URL",
                    nargs="?",
                    help=("url of song, album or playlist from spotify."),
                    )


def parse_cmd():
    args = parser.parse_args()
    if args.user in ['current', 'current-playing']:
        args.URL = 'current'
    elif args.user in ['play', 'playlist']:
        args.URL = 'play'
    elif args.user in ['album']:
        args.URL = 'album'
    if args.config == 'edit':
        create_config()
    elif args.config in ["reset", "r"]:
        create_config(config_exists = False)
    elif args.config in ["open", "o"]:
        open_config()
    return args.URL, args.directory, args.force

def input_taker(config, key, question, string = True):
    print(question)
    print(f"[SAVED]: {str(config[key])[:50]}")
    if ans := input():
        config[key] = ans if string else bool(int(ans))

def check_config():
    global CONFIG_FILE
    OS_CONFIG = os.environ.get("APPDATA") if os.name == "nt" else os.path.join(os.environ["HOME"], ".config")
    CONFIG_PATH = os.path.join(OS_CONFIG, "syrics")
    CONFIG_FILE = os.path.join(CONFIG_PATH, "config.json")
    if not os.path.exists(CONFIG_FILE):
        os.makedirs(CONFIG_PATH, exist_ok=True)
        create_config(config_exists=False)
    return CONFIG_FILE


def create_config(config_exists = True):
    print("Editing Config File...")
    print("To keep the current value just press enter!")
    config = {
        'sp_dc': "", 
        'download_path': "downloads", 
        'create_folder': True, 
        'album_folder_name': "{name} - {artists}", 
        'play_folder_name': "{name} - {owner}", 
        'file_name': "{track_number}. {name}",
        'synced_lyrics': True, 
        'force_download': False
    }
    if config_exists:
        with open(CONFIG_FILE) as f:
            config = json.load(f)
    input_taker(config, 'sp_dc', "Enter the sp_dc: ")
    input_taker(config, 'download_path', "Enter the download path: ")
    input_taker(config, 'create_folder', "create folder for album/playlists: (0: False, 1: True)", string=False)
    input_taker(config, 'album_folder_name', "Enter the album folder naming format: ")
    input_taker(config, 'play_folder_name', "Enter the playlist folder naming format: ")
    input_taker(config, 'file_name', "Enter the file naming format: ")
    input_taker(config, 'synced_lyrics', "Get synced lyrics: (0: False, 1: True)", string=False)
    input_taker(config, 'force_download', "Skip check for if it already exists: (0: False, 1: True)", string=False)
    with open(CONFIG_FILE, "w+") as f:
        f.write(json.dumps(config, indent=4))
    print("config sucessfully setup, run the program again.")
    sys.exit(1)

def open_config():
    if os.name == "nt":
        subprocess.call(["notepad", CONFIG_FILE])
    else:
        subprocess.call(["nano", CONFIG_FILE])
    sys.exit(1)


from setuptools import setup, find_packages

pkg_name = "syrics"


def read_file(fname):
    with open(fname, "r") as f:
        return f.read()


requirements = [
    "requests",
    "spotipy",
    "tqdm",
    "tinytag",
]

setup(
    name=pkg_name,
    version="0.1.2.4",
    author="Akash R Chandran",
    author_email="chandranrakash@gmail.com",
    description="A command line tool to fetch lyrics from spotify and save it to lrc file. It can fetch both synced and unsynced lyrics from spotify. ",
    long_description=read_file("README.md"),
    long_description_content_type="text/markdown",
    url="https://github.com/akashrchandran/syrics",
    install_requires=requirements,
    entry_points={
        "console_scripts": [
            "syrics = syrics:main",
        ],
    },
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)

# rm -f dist/*
# python3 setup.py sdist bdist_wheel
# twine upload dist/*


#!/usr/bin/env python3
import json
import os
import re

from tinytag import TinyTag
from tqdm import tqdm

from syrics.api import Spotify
from syrics.cli import parse_cmd, check_config
from syrics.exceptions import CorruptedConfig

logo = '''
     _______.____    ____ .______       __    ______     _______.
    /       |\   \  /   / |   _  \     |  |  /      |   /       |
   |   (----` \   \/   /  |  |_)  |    |  | |  ,----'  |   (----`
    \   \      \_    _/   |      /     |  | |  |        \   \    
.----)   |       |  |     |  |\  \----.|  | |  `----.----)   |   
|_______/        |__|     | _| `._____||__|  \______|_______/    
                                                                 

'''


def get_album_tracks(album_id: str):
    album_data = client.album(album_id)
    album_data['artists'] = ','.join([artist['name'] for artist in album_data['artists']])
    album_folder = rename_using_format(config['album_folder_name'], album_data)
    print(f"> Album: {album_data['name']}")
    print(f"> Artist: {album_data['artists']}")
    print(f"> Songs: {album_data['total_tracks']} Tracks", end='\n\n')
    return client.album_tracks(album_id, album_data['total_tracks']), album_folder


def get_playlist_tracks(playlist_id: str):
    play_data = client.playlist(playlist_id)
    play_data['owner'] = play_data['owner']['display_name']
    play_data['total_tracks'] = play_data['tracks']['total']
    play_data['collaborative'] = '[C]' if play_data['collaborative'] else ''
    play_folder = rename_using_format(config['play_folder_name'], play_data)
    print(f"> Playlist: {play_data['name']}")
    print(f"> Owner: {play_data['owner']}")
    print(f"> Songs: {play_data['total_tracks']} Tracks", end='\n\n')
    return client.playlist_tracks(playlist_id, play_data['tracks']['total']), play_folder


def format_lrc(lyrics_json, track_data):
    lyrics = lyrics_json['lyrics']['lines']
    minutes, seconds = divmod(int(track_data["duration_ms"]) / 1000, 60)
    lrc = [
        f'[ti:{track_data["name"]}]',
        f'[al:{track_data["album_name"]}]',
        f'[ar:{track_data["artist"]}]',
        f'[length: {minutes:0>2.0f}:{seconds:05.2f}]',
    ]
    for lines in lyrics:
        if lyrics_json['lyrics']['syncType'] == 'UNSYNCED' or not config['synced_lyrics']:
            lrc.append(lines['words'])
        else:
            duration = int(lines['startTimeMs'])
            minutes, seconds = divmod(duration / 1000, 60)
            lrc.append(f'[{minutes:0>2.0f}:{seconds:05.2f}] {lines["words"]}')
    return '\n'.join(lrc)


def sanitize_track_data(track_data: dict):
    album_data = track_data['album']
    artist_data = track_data['artists']
    del track_data['album']
    del track_data['artists']
    track_data['album_name'] = album_data['name']
    track_data['release_date'] = album_data['release_date']
    track_data['total_tracks'] = str(album_data['total_tracks']).zfill(2)
    track_data['track_number'] = str(track_data['track_number']).zfill(2)
    track_data['album_artist'] = ','.join(
        [artist['name'] for artist in album_data['artists']])
    track_data['artist'] = ','.join([artist['name'] for artist in artist_data])
    track_data['explicit'] = '[E]' if track_data['explicit'] else ''


def save_lyrics(lyrics, path):
    with open(path, "w+", encoding='utf-8') as f:
        f.write(lyrics)


def rename_using_format(string: str, data: dict):
    matches = re.findall('{(.+?)}', string)
    for match in matches:
        word = '{%s}' % match
        string = string.replace(word, str(data[match]))
    return re.sub(r'[\\/*?:"<>|]',"", string)

def chunk(lst, n):
    for i in range(0, len(lst), n):
        yield lst[i:i + n]

def download_lyrics(track_ids: list, folder: str = None):
    unable = []
    if folder:
        folder = f"{config['download_path']}/{folder}"
        if config['create_folder'] and not os.path.exists(folder) and config.get('force_download'):
            os.makedirs(folder, exist_ok=True)
        else:
            print("The album/playlist was already downloaded..skipping")
            exit(0)
    else:
        folder = config['download_path']
    tracks_data = []
    for trackid in chunk(track_ids, 50):
        tracks_data += client.tracks(trackid)['tracks']
    for track in tqdm(tracks_data):
        sanitize_track_data(track)
        lyrics_json = client.get_lyrics(track['id'])
        if not lyrics_json:
            unable.append(track['name'])
            continue
        file_name = f"{folder}/{rename_using_format(config['file_name'], track)}.lrc"
        save_lyrics(format_lrc(lyrics_json, track), path=file_name)
    return unable

def fetch_files(path: str):
    files = [tracks for tracks in os.listdir(path) if re.search("\.(flac|mp3|wav|ogg|opus|m4a|aiff)$", tracks)]
    unable = []
    for files in tqdm(files):
            tag = TinyTag.get(os.path.join(path, files))
            query = client.search(q=f"track:{tag.title} album:{tag.album}", type="track", limit=1)
            if track := query['tracks']['items']:
                sanitize_track_data(track[0])
                file_name = f"{os.path.splitext(tag._filehandler.name)[0]}.lrc"
                lyrics_json = client.get_lyrics(track[0]['id'])
                if not lyrics_json:
                    unable.append(tag.title)
                    continue
                if not os.path.exists(file_name) and config.get('force_download'):
                    save_lyrics(format_lrc(lyrics_json, track[0]), path=file_name)
            else:
                unable.append(tag.title)
    return unable

def initial_checks():
    global client, cmd_url, config
    CONFIG_FILE = check_config()
    cmd_url, directory, force = parse_cmd()
    try:
        with open(CONFIG_FILE) as f:
            config = json.load(f)
    except Exception as e:
        raise CorruptedConfig("Config file seems corrupted, run syrics -c reset") from e
    config['download_path'] = directory or config['download_path']
    config['force_download'] = force or config['force_download']
    client = Spotify(config['sp_dc'])
    if cmd_url == 'current':
        cmd_url = client.get_current_song( )['item']['external_urls']['spotify']
    elif cmd_url == 'play':
        cmd_url = client.select_user_playlist()['external_urls']['spotify']
    elif cmd_url == 'album':
        cmd_url = client.select_user_album()['external_urls']['spotify']

def main():
    initial_checks()
    if config['download_path'] and not os.path.exists(config['download_path']):
        os.mkdir(config['download_path'])
    print(logo, end='\n\n')
    account = client.get_me()
    print("Successfully Logged In as:")
    print("Name: " + account["display_name"])
    print("Country: " + account["country"], end='\n\n')
    link = cmd_url or input("Enter Link: ")
    if 'spotify' in link:
        if 'album' in link:
            track_ids, folder = get_album_tracks(link)
        elif 'playlist' in link:
            track_ids, folder= get_playlist_tracks(link)
        elif 'track' in link:
            track_ids = [link]
            folder = None
        else:
            print("Enter valid url!")
            exit(0)
        unable = download_lyrics(track_ids, folder)
    else:
        unable = fetch_files(link)
    if unable:
        print("\nsome tracks does not have lyrics, so skipped:")
        for tracks in unable:
            print(tracks)
